#
# AWS Secrets Manager (v1.1)
#
# aws_secret input fields:
#
#  * secret_id
#  * version_stage (optional)
#  * alias (optional) - use this name in the vault instead of [secret_id]
#  * profile (optional)
#
# aws_secret output fields:
#
#  * secret - the secret string value, empty on error
#  * response_status - the HTTP response code from AWS, 0 on error
#  * response_code - internal network error code, non zero on error
#  * response_body - the HTTP response body, empty on error
#
# aws_secret_vault output fields:
#
#  * [secret_id] - the last successful secret string value
#

import crypto;
import http;
import json;
import kvstore;
import utils;

sub vcl_init
{
	# From init
	#new aws_sign = kvstore.init(scope = REQUEST);

	# Init the secret storage
	new aws_secret = kvstore.init(scope = REQUEST);
	new aws_secret_vault = kvstore.init();
}

sub aws_get_secret
{
	# Do we have a secret_id?
	if (!aws_secret.get("secret_id")) {
		aws_secret.init(scrub = true);
		call aws_sign_finish;
		return (fail("AWS secret id not found. Required: secret_id"));
	}

	# Setup
	aws_sign.set("_secret_id", aws_secret.get("alias", aws_secret.get("secret_id")));

	if (aws_secret.get("profile")) {
		aws_sign.set("profile", aws_secret.get("profile"));
	}

	# Start the AWS v4 secretsmanager signed request
	call aws_sign_init;

	if (aws_secret.get("version_stage")) {
		aws_sign.set("_payload", {"
		{
		  "SecretId": "} + json.stringify(aws_secret.get("secret_id")) + {",
		  "VersionStage": "} + json.stringify(aws_secret.get("version_stage")) + {"
		}
		"});
	} else {
		aws_sign.set("_payload", {"
		{
		  "SecretId": "} + json.stringify(aws_secret.get("secret_id")) + {"
		}
		"});
	}

	aws_sign.set("_host", "secretsmanager." + aws_sign.get("_region") + ".amazonaws.com");
	aws_sign.set("_datestamp", utils.time_format("%Y%m%d"));
	aws_sign.set("_signed_headers", "host;x-amz-content-sha256;x-amz-date;x-amz-target");
	aws_sign.set("_scope", aws_sign.get("_datestamp") + "/" +
		aws_sign.get("_region") + "/secretsmanager/aws4_request");

	http.init(8, LOW);
	http.req_set_method(8, "POST");
	if (aws_secret.get("endpoint")) {
		http.req_set_url(8, aws_secret.get("endpoint"));
	} else {
		http.req_set_url(8, "https://" + aws_sign.get("_host") + "/");
	}
	http.req_set_sparam(8, "POSTFIELDS", aws_sign.get("_payload"));
	http.req_set_header(8, "Host", aws_sign.get("_host"));
	http.req_set_header(8, "Content-Type", "application/x-amz-json-1.1");

	http.req_set_header(8, "x-amz-target", "secretsmanager.GetSecretValue");
	http.req_set_header(8, "x-amz-date", utils.time_format("%Y%m%dT%H%M%SZ"));
	http.req_set_header(8, "x-amz-content-sha256",
		crypto.hex_encode(crypto.hash(sha256, aws_sign.get("_payload"))));

	aws_sign.set("_canonical_request",
		crypto.hex_encode(crypto.hash(sha256,
			"POST" + utils.newline() +
			"/" + utils.newline() +
			"" + utils.newline() +
			"host:" + http.req_get_header(8, "Host") + utils.newline() +
			"x-amz-content-sha256:" + http.req_get_header(8, "x-amz-content-sha256") + utils.newline() +
			"x-amz-date:" + http.req_get_header(8, "x-amz-date") + utils.newline() +
			"x-amz-target:" + http.req_get_header(8, "x-amz-target") + utils.newline() +
			utils.newline() +
			aws_sign.get("_signed_headers") + utils.newline() +
			http.req_get_header(8, "x-amz-content-sha256"))
		));

	aws_sign.set("_signature",
		crypto.hex_encode(crypto.hmac(sha256,
			crypto.hmac(sha256,
			crypto.hmac(sha256,
			crypto.hmac(sha256,
			crypto.hmac(sha256,
			crypto.blob("AWS4" + aws_sign.get("_secret_key")),
				aws_sign.get("_datestamp")),
				aws_sign.get("_region")),
				"secretsmanager"),
				"aws4_request"),
					"AWS4-HMAC-SHA256" + utils.newline() +
					http.req_get_header(8, "x-amz-date") + utils.newline() +
					aws_sign.get("_scope") + utils.newline() +
					aws_sign.get("_canonical_request"))
		));

	http.req_set_header(8, "Authorization",
		"AWS4-HMAC-SHA256 " +
		"Credential=" + aws_sign.get("_access_key") + "/" +
		aws_sign.get("_scope") + ", " +
		"SignedHeaders=" + aws_sign.get("_signed_headers") + ", " +
		"Signature=" + aws_sign.get("_signature"));

	# Send the request
	http.req_send(8);
	http.resp_wait(8);

	# Set the response fields
	aws_secret.init(scrub = true);
	aws_secret.set("response_status", http.resp_get_status(8));
	aws_secret.set("response_code", http.resp_get_errorcode(8));
	aws_secret.set("response_body", http.resp_get_body(8));

	# Find the SecretString in the payload
	if (http.resp_get_status(8) == 200) {
		json.parse(http.resp_get_body(8));

		if (json.get("SecretString")) {
			aws_secret.set("secret", json.get("SecretString"));
			aws_secret_vault.set(aws_sign.get("_secret_id"), json.get("SecretString"));
		} else if (json.get("SecretBinary")) {
			aws_secret.set("secret", json.get("SecretBinary"));
			aws_secret_vault.set(aws_sign.get("_secret_id"), json.get("SecretBinary"));
		}
	}

	# Done
	http.finish(8);
	call aws_sign_finish;
}

# EOF
