#
# Cluster v2.1
#

import udo;
import std;
import kvstore;
import headerplus;
import accounting;

sub vcl_init {
	new cluster = udo.director();
	new cluster_opts = kvstore.init();
	new cluster_stats = kvstore.init();
	new cluster_req = kvstore.init(scope = REQUEST);

	# Available options
	cluster_opts.set("token", "Warning, default token!");
	# cluster_opts.set("primaries", "1");
	# cluster_opts.set("trace", "false");
	# cluster_opts.set("fallback", "3");

	cluster_stats.counter("error_token", 0, varnishstat = true, comment = "Bad tokens received");
	cluster_stats.counter("error_fallback_limit", 0, varnishstat = true, comment = "Cluster fallback limit exceeded");
	cluster_stats.counter("error_unhealthy", 0, varnishstat = true, comment = "No healthy nodes in the cluster");
	cluster_stats.counter("skipped", 0, varnishstat = true, comment = "Requests not Self-Routed due to X-Cluster-Skip being set");
	cluster_stats.counter("passed", 0, varnishstat = true, comment = "Requests not Self-Routed due to being a PASS");
	cluster_stats.counter("hitmiss", 0, varnishstat = true, comment = "Requests not Self-Routed due to being a Hit-For-Miss");
	cluster_stats.gauge("self_identified", 0, varnishstat = true, comment = "The node has identified itself");

	accounting.create_namespace("cluster");
}

sub cluster_clean {
	headerplus.init_req();
	headerplus.delete_regex("^X-Cluster-");
	headerplus.write();
}

sub cluster_trace {
	# Trace: Enabled by setting the "trace" option to true.
	headerplus.init_resp();
	headerplus.set("X-Cluster-Trace", server.identity);
	if (bereq.uncacheable) {
		headerplus.append("X-Cluster-Trace", "PASS", "->");
	} else if (bereq.is_hitmiss) {
		headerplus.append("X-Cluster-Trace", "HITMISS", "->");
	} else {
		headerplus.append("X-Cluster-Trace", "MISS", "->");
	}
	if (bereq.retries > 0) {
		headerplus.append("X-Cluster-Trace", "RETRY(" + bereq.retries + ")", "->");
	}
	if (beresp.was_304) {
		headerplus.append("X-Cluster-Trace", "304", "->");
	}
	if (bereq.http.X-Cluster-Error == "true") {
		headerplus.append("X-Cluster-Trace", "ERROR", "->");
	} else if (bereq.backend == cluster.backend()) {
		headerplus.append("X-Cluster-Trace", beresp.http.X-Cluster-Trace, ", ");
	} else {
		headerplus.append("X-Cluster-Trace", bereq.backend, ", ");
	}
	headerplus.write();
}

sub cluster_accounting_backend_response {
	# Accounting: Add accounting keys based on whether the backend response
	# is was received from origin or another cluster node.
	if (cluster_req.get("skip-accounting")) {
		return;
	}
	if (!accounting.get_namespace()) {
		accounting.set_namespace("cluster");
	}
	if (bereq.backend == cluster.backend()) {
		accounting.add_keys("cluster_backend_response");
	} else {
		accounting.add_keys("origin_backend_response");
	}
}

sub cluster_accounting_deliver {
	# Accounting: Add accounting keys based on whether the response is being
	# delivered to a client or another cluster node.
	if (req.http.X-Cluster-Skip-Accounting == "true") {
		return;
	}
	if (!accounting.get_namespace()) {
		accounting.set_namespace("cluster");
	}
	if (req.http.X-Cluster-Token) {
		accounting.add_keys("cluster_deliver");
	} else {
		accounting.add_keys("client_deliver");
	}
}

sub vcl_recv {
	unset req.http.X-Cluster-Skip;
	unset req.http.X-Cluster-Skip-Accounting;

	if (req.http.X-Cluster-Token) {
		# Token validation: The token is used to verify that a request
		# is coming from another cluster member.
		if(req.http.X-Cluster-Token != cluster_opts.get("token")) {
			unset req.http.X-Cluster-Token;
			std.log("Cluster: Error, bad token");
			cluster_stats.counter("error_token", 1);
			return (synth(403, "Bad cluster token"));
		}
		if (cluster.self_identify(req.http.X-Cluster-Identifier)) {
			# We only end up here if the director has not yet identified itself,
			# and this request is a self-identification loop.
			set req.http.X-Cluster-Skip-Accounting = "true";
			return (synth(242, "Self-Identified"));
		}
		if (req.http.X-Cluster-Is-Bgfetch == "true") {
			# Incoming bgfetch: When a cluster node self-routes a
			# bgfetch, we don't want to serve a stale object.
			set req.grace = 0s;
		}
		if (req.esi_level > 0) {
			# ESI: Enable self-routing for subrequests
			call cluster_clean;
		}
	}
}

sub vcl_backend_fetch {
	if (cluster.is_identified()) {
		# Multiple primaries: Use the random subtype to balance load
		# over all primaries. Safe to do only after self-identification.
		cluster.set_subtype(random, std.integer(cluster_opts.get("primaries"), 1));
	}

	if (bereq.retries == 0 && bereq.backend) {
		# Backend hint: Save req.backend_hint
		cluster_req.set_backend("hint", bereq.backend);
	}

	if (bereq.http.X-Cluster-Skip-Accounting == "true") {
		# Accounting, Remember skip
		cluster_req.set("skip-accounting", "true");
	}

	if (bereq.http.X-Cluster-Skip == "true") {
		std.log("Cluster: Not self-routing, skipped by VCL");
		cluster_stats.counter("skipped", 1);
		cluster_req.set("skip", "true");
	} else if (bereq.uncacheable) {
		std.log("Cluster: Not self-routing, request is a PASS");
		cluster_stats.counter("passed", 1);
	} else if (bereq.is_hitmiss) {
		std.log("Cluster: Not self-routing, fetch is a Hit-For-Miss");
		cluster_stats.counter("hitmiss", 1);
	} else if (bereq.backend != cluster.backend() && bereq.retries > 0) {
		std.log("Cluster: Not self-routing, this is an origin retry");
	} else if (bereq.retries > std.integer(cluster_opts.get("fallback"), 3)) {
		std.log("Cluster: Not self-routing, fallback limit reached");
		cluster_stats.counter("error_fallback_limit", 1);
	} else if (cluster.self_is_next(std.integer(cluster_opts.get("primaries"), 1))) {
		std.log("Cluster: Not self-routing, this is a primary node");
	} else if (bereq.http.X-Cluster-Token) {
		std.log("Cluster: Not self-routing, this is a self-routed request");
	} else if (!std.healthy(cluster.backend())) {
		std.log("Cluster: Not self-routing, no healthy nodes in the cluster");
		cluster_stats.counter("error_unhealthy", 1);
	} else {
		std.log("Cluster: Self-routing to primary node");
		call cluster_clean;
		set bereq.http.X-Cluster-Identifier = cluster.get_identifier();
		set bereq.http.X-Cluster-Token = cluster_opts.get("token");
		set bereq.http.X-Cluster-Is-Bgfetch = bereq.is_bgfetch;
		set bereq.backend = cluster.backend();
		return (fetch);
	}

	call cluster_clean;

	if (bereq.retries > 0 && bereq.backend == cluster.backend()) {
		# Backend hint: Restore req.backend_hint
		set bereq.backend = cluster_req.get_backend("hint");
	}
}

sub vcl_backend_error {
	if (bereq.backend == cluster.backend()) {
		# Automatic fallback: Always retry when a cluster fetch fails.
		unset bereq.http.X-Cluster-Token;
		return (retry);
	}

	set bereq.http.X-Cluster-Error = "true";
	call cluster_trace;
	call cluster_accounting_backend_response;
}

sub vcl_backend_response {
	if (bereq.backend == cluster.backend()) {
		if (beresp.status == 242 && beresp.reason == "Self-Identified") {
			# The request looped, and this node has identified itself.
			# Retry to self-route the request or go to origin.
			std.log("Cluster: Node self-identified");
			cluster_stats.gauge("self_identified", 1);
			return (retry);
		}
	}

	call cluster_trace;
	call cluster_accounting_backend_response;
}

sub vcl_deliver {
	if (obj.hits > 0 && resp.http.X-Cluster-Trace !~ "->ERROR$") {
		if (resp.is_streaming) {
			set resp.http.X-Cluster-Trace = server.identity + "->HITSTREAM";
		} else {
			set resp.http.X-Cluster-Trace = server.identity + "->HIT";
		}
	}

	if (req.http.X-Cluster-Token) {
		# ESI: Disable ESI resolution for cluster-intenal responses
		set resp.do_esi = false;
	} else {
		if (cluster_opts.get("trace") != "true") {
			unset resp.http.X-Cluster-Trace;
		}
	}

	call cluster_accounting_deliver;
}

sub vcl_synth {
	set resp.http.X-Cluster-Trace = server.identity + "->SYNTH";

	if (req.http.X-Cluster-Token) {
		if (resp.status == 242 && resp.reason == "Self-Identified") {
			return (deliver);
		}
	} else {
		if (cluster_opts.get("trace") != "true") {
			unset resp.http.X-Cluster-Trace;
		}
	}

	call cluster_accounting_deliver;
}
